'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.queryASTToSqlAST = queryASTToSqlAST;
exports.getGraphQLType = getGraphQLType;

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function queryASTToSqlAST(ast) {
  const usedTableAliases = new Set();

  const sqlAST = {};
  _assert2.default.equal(ast.fieldASTs.length, 1, 'We thought this would always have a length of 1. FIX ME!!');

  const queryAST = ast.fieldASTs[0];

  const parentType = ast.parentType;
  getGraphQLType(queryAST, parentType, sqlAST, ast.fragments, usedTableAliases);
  return sqlAST;
}

function getGraphQLType(queryASTNode, parentTypeNode, sqlASTNode, fragments, usedTableAliases) {
  const fieldName = queryASTNode.name.value;

  let field = parentTypeNode._fields[fieldName];

  let grabMany = false;

  let gqlType = stripNonNullType(field.type);

  if (gqlType.constructor.name === 'GraphQLList') {
    gqlType = gqlType.ofType;
    grabMany = true;
  }

  if (/Connection$/.test(gqlType.name) && gqlType.constructor.name === 'GraphQLObjectType' && gqlType._fields.edges) {
    grabMany = true;

    const stripped = stripRelayConnection(field, queryASTNode);

    gqlType = stripped.gqlType;
    queryASTNode = stripped.queryASTNode;

    sqlASTNode.relayPaging = true;
  }

  const config = gqlType._typeConfig;

  if (gqlType.constructor.name === 'GraphQLObjectType' && config.sqlTable) {
    handleTable(sqlASTNode, queryASTNode, field, gqlType, fragments, usedTableAliases, grabMany);
  } else if (field.sqlColumn || !field.resolve) {
    sqlASTNode.type = 'column';
    sqlASTNode.name = field.sqlColumn || field.name;
    sqlASTNode.fieldName = field.name;
  } else if (field.sqlDeps) {
    sqlASTNode.type = 'columnDeps';
    sqlASTNode.name = field.sqlDeps;
  } else {
    sqlASTNode.type = 'noop';
  }
}

function handleTable(sqlASTNode, queryASTNode, field, gqlType, fragments, usedTableAliases, grabMany) {
  const config = gqlType._typeConfig;

  sqlASTNode.type = 'table';
  sqlASTNode.name = config.sqlTable;

  sqlASTNode.as = makeUnique(usedTableAliases, field.name);

  if (queryASTNode.arguments.length) {
    const args = sqlASTNode.args = {};
    for (let arg of queryASTNode.arguments) {
      args[arg.name.value] = arg.value.value;
    }
  }

  sqlASTNode.fieldName = field.name;
  sqlASTNode.grabMany = grabMany;

  if (field.where) {
    sqlASTNode.where = field.where;
  }
  if (field.sqlJoin) {
    sqlASTNode.sqlJoin = field.sqlJoin;
  }
  if (field.joinTable) {
    sqlASTNode.sqlJoins = field.sqlJoins;
    sqlASTNode.joinTable = field.joinTable;
    sqlASTNode.joinTableAs = makeUnique(usedTableAliases, field.joinTable);
  }

  const children = sqlASTNode.children = [];

  if (!config.uniqueKey) {
    throw new Error(`You must specify the "uniqueKey" on the GraphQLObjectType definition of ${ config.sqlTable }`);
  }
  if (typeof config.uniqueKey === 'string') {
    children.push({
      type: 'column',
      name: config.uniqueKey,
      fieldName: config.uniqueKey
    });
  } else if (Array.isArray(config.uniqueKey)) {
    children.push({
      type: 'composite',
      name: config.uniqueKey,
      fieldName: config.uniqueKey.join('#')
    });
  }

  if (queryASTNode.selectionSet) {
    for (let selection of queryASTNode.selectionSet.selections) {
      switch (selection.kind) {
        case 'Field':
          growNewTreeAndAddToChildren(children, selection, gqlType, fragments, usedTableAliases);
          break;

        case 'InlineFragment':
          if (selection.typeCondition.name.value === gqlType.name) {
            for (let fragSelection of selection.selectionSet.selections) {
              growNewTreeAndAddToChildren(children, fragSelection, gqlType, fragments, usedTableAliases);
            }
          }
          break;

        case 'FragmentSpread':
          const fragmentName = selection.name.value;
          const fragment = fragments[fragmentName];

          if (fragment.typeCondition.name.value === gqlType.name) {
            for (let fragSelection of fragment.selectionSet.selections) {
              growNewTreeAndAddToChildren(children, fragSelection, gqlType, fragments, usedTableAliases);
            }
          }
          break;
        default:
          throw new Error('Unknown selection kind: ' + selection.kind);
      }
    }
  }
}

function stripRelayConnection(field, queryASTNode) {
  const gqlType = field.type._fields.edges.type.ofType._fields.node.type;

  const args = queryASTNode.arguments;

  const edges = queryASTNode.selectionSet.selections.find(selection => selection.name.value === 'edges');
  if (edges) {
    queryASTNode = edges.selectionSet.selections.find(selection => selection.name.value === 'node') || {};
  } else {
    queryASTNode = {};
  }

  queryASTNode.arguments = args;
  return { gqlType, queryASTNode };
}

function growNewTreeAndAddToChildren(children, selection, graphQLType, fragments, usedTableAliases) {
  const newNode = {};
  children.push(newNode);
  getGraphQLType(selection, graphQLType, newNode, fragments, usedTableAliases);
}

function stripNonNullType(type) {
  return type.constructor.name === 'GraphQLNonNull' ? type.ofType : type;
}

function makeUnique(usedNames, name) {
  if (usedNames.has(name)) {
    name += '$';
  }
  usedNames.add(name);
  return name;
}